/*
  adapter_pci - Retrieve VPD for PCI adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter_pci.c,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/

#define _GNU_SOURCE

#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <stdio.h>

#include "lsvpd.h"
#include "lsvpd_lib_dir.h"
#include "util.h"
#include "init.h"

#include "adapter_pci_vpd_capabilities.h"
#include "adapter_pci_vpd_parse.h"
#include "adapter_pci_lookup.h"
#include "adapter.h"
#include "adapter_pci.h"

/********************************************************************/

#define ADAPTER_PCI_IDS_FILENAME "pci.ids"

/********************************************************************/

adapter_pci_get_field_value_t * adapter_pci_get_field_value_function = NULL;

char *
adapter_pci_get_field_value(const char * node,
			    const char * field)
{
	char * ret = NULL;

	if (NULL != adapter_pci_get_field_value_function)
		ret = (*adapter_pci_get_field_value_function) (node, field);
	return ret;
}

/********************************************************************/

static vpd_list_t *
pci_manufacture_vpd(const bus_addr_t bus_addr,
		    const node_t subtype,
		    const char * node)
{
	vpd_list_t * ret;
	const char * ds;
	char * source_node = NULL;
	char * vendor = NULL;
	char * device = NULL;
	char * subvendor = NULL;
	char * subdevice = NULL;
	bool have_subs;
	char * t;
	char * mf;
	char * tm;
	char * submf;
	char * subtm;

	ret = vpd_list_new();

	ds = adapter_get_ds(ADAPTER_TYPE_PCI, bus_addr, subtype);

	if (NULL == ds)
		goto bail;

	source_node = adapter_get_source_node(ADAPTER_TYPE_PCI,
					      bus_addr, subtype);

	if (NULL == source_node)
		goto bail;

	vendor = adapter_pci_get_field_value(source_node, "vendor");
	device = adapter_pci_get_field_value(source_node, "device");

	if ((NULL == vendor) || (NULL == device))
		goto bail;

	/* This is to workaround a broken adapter.  We're much more
	 * likely to see the Ethernet adapter from AMD (1022) than the
	 * display adapter from Trident (1023). */
	if ((0 == strcmp(vendor, "1023")) &&
	    (0 == strcmp(device, "2000")))
		vendor[3] = '2';

	subvendor = adapter_pci_get_field_value(source_node,
						"subsystem_vendor");
	subdevice = adapter_pci_get_field_value(source_node,
						"subsystem_device");

	/* Add DS field, modifying with subsystem info if possible. */
	have_subs = (NULL != subvendor) && (NULL != subdevice) &&
		(4 == strlen(subvendor)) && (4 == strlen(subdevice));
	if (have_subs &&
	    (-1 != asprintf(&t, "%s (%.2s%.2s%.2s%.2s)", ds,
			    &subvendor[2], &subvendor[0],
			    &subdevice[2], &subdevice[0]))) {
		vpd_list_maybe_append_key_string(&ret, "DS", t);
		free(t);
	} else
		vpd_list_maybe_append_key_string(&ret, "DS", ds);

	t = lsvpd_relative_filename(lsvpd_lib_dir, ADAPTER_PCI_IDS_FILENAME);
	if (NULL == t)
		goto bail;

	/* Convert hex values into friendly strings. */
	if (!adapter_pci_lookup(t,
				vendor, device, subvendor, subdevice,
				&mf, &tm, &submf, &subtm)) {
		free(t);
		goto bail;
	}

	/* If subtm was found, use subtm and submf instead. */
	if (NULL != subtm) {
		free(tm);
		free(mf);
		tm = subtm;
		mf = submf;
	} else if (NULL != submf)
		/* Free unused submf. */
		free(submf);

	/* If necessary, try suitable defaults. */
	if ((NULL == mf) &&
	    (-1 == asprintf(&mf, "Unknown (0x%s)", vendor)))
		mf = NULL;
	if ((NULL == tm) &&
	    (-1 == asprintf(&tm, "Unknown (0x%s)", device)))
		tm = NULL;

	vpd_list_maybe_append_key_string(&ret, "MF", mf);
	vpd_list_maybe_append_key_string(&ret, "TM", tm);

	/* Construct CD field from subvendor and subdevice. */
	if (have_subs &&
	    (-1 != asprintf(&t, "%s%s", subvendor, subdevice))) {
		vpd_list_maybe_append_key_string(&ret, "CD", t);
		free(t);
	}

	free(mf);
	free(tm);

done:
	free(source_node);
	free(vendor);
	free(device);
	free(subvendor);
	free(subdevice);

	return ret;
bail:
	vpd_list_free(ret);
	ret = NULL;
	goto done;
}

/********************************************************************/

static bool
pci_addr_sanity_check(const bus_addr_t bus_addr)
{
	return ((12 == strlen(bus_addr)) &&
		(':' == bus_addr[4]) &&
		(':' == bus_addr[7]) &&
		('.' == bus_addr[10])) ||
		((7 == strlen(bus_addr)) &&
		 (':' == bus_addr[2]) &&
		 ('.' == bus_addr[5]));
}

/* Can use NULL below and put default functionality inside main function. */
adapter_pci_get_config_file_t * adapter_pci_get_config_file_function = NULL;

char *
adapter_pci_get_config_file(const bus_addr_t bus_addr)
{
	char * ret;

	ret = NULL;

	if (NULL != adapter_pci_get_config_file_function)
		ret = (*adapter_pci_get_config_file_function)(bus_addr);
	else if (pci_addr_sanity_check(bus_addr)) {
		/* The big question here is how much parsing to bother
		 * with.  This approach card-codes as much as possible
		 * without really doing any parsing.  It works.  If
		 * there were another place where we actually had to
		 * pull these things apart, we'd parse it properly,
		 * remembering if there was a PCI domain in the parse
		 * result.  Bear in mind that this is fall-back code
		 * for Linux 2.4. */

		/* Try "/proc/bus/pci/[dom:]bus/dev.fun". */
		if (-1 != asprintf(&ret, "/proc/bus/pci/%.*s/%s",
				   (':' == bus_addr[4]) ? 7 : 2,
				   bus_addr,
				   bus_addr + ((':' == bus_addr[4]) ? 8 : 3))) {
			if (0 != access(ret, R_OK)) {
				/* Doesn't exist.  Try again.  This
				 * time, if there is a PCI domain and
				 * it is 0000 then try without it. */
				free(ret);
				ret = NULL;
				
				if (0 == strncmp(bus_addr, "0000:", 5)) {
					if (-1 != asprintf(&ret,
							   "/proc/bus/pci/%.2s/%s",
							   bus_addr + 5,
							   bus_addr + 8)) {
						if (0 != access(ret, R_OK)) {
							free(ret);
							ret = NULL;
						}
					} else
						ret = NULL;
				}
			}
		}
	} else
		ret = NULL;

	return ret;
}

static void
pci22_vpd_write_tempfile(const char * node, const growbuf * buf)
{
	char * t;

	t = lsvpd_relative_filename(node, LSVPD_TMP_SUBDIR);
	if (NULL != t) {
		if (lsvpd_mkdir(t))
			lsvpd_file_relative_write(t, "pci-2.2.vpd",
						  growbuf_data(buf),
						  growbuf_length(buf));
		free(t);
	}
}

static vpd_list_t *
pci_retrieve_pci22_vpd(const bus_addr_t bus_addr,
		       const node_t subtype,
		       const char * node)
{
	vpd_list_t * ret;
	growbuf * buf;
	char * t;
	int lockfd;
	char * config_file;

	ret = NULL;

	if (-1 != asprintf(&t, "%s/pci-%s.lock", lsvpd_db_lock_dir, bus_addr)) {
		lockfd = lsvpd_lock(t);
		if (-1 != lockfd) {
			config_file = adapter_pci_get_config_file(bus_addr);
			if (NULL != config_file) {
				buf = adapter_pci_vpd_capabilities(config_file);
				if (NULL != buf) {
					pci22_vpd_write_tempfile(node, buf);
					ret = adapter_pci_vpd_parse(buf);
					growbuf_free(buf);
				}
				
				free(config_file);
			}
			lsvpd_unlock(lockfd);
		}
		free(t);
	}

	return ret;
}

/********************************************************************/

static adapter_get_ds_function_t * pci_get_ds_basic = NULL;

static char *
pci_get_ds_function(const node_t bus_type,
		    const bus_addr_t bus_addr,
		    const node_t subtype)
{
	char * ret;
	char * source_node = NULL;
	char * class = NULL;
	char * t = NULL;
	char v[5];
	char * c;
	char * x1;
	char * x2;

	ret = NULL;

	if ((0 != strcmp(ADAPTER_SUBTYPE_GENPCI, subtype)) &&
	    (NULL != pci_get_ds_basic)) {
		ret = (*pci_get_ds_basic)(bus_type, bus_addr, subtype);
		goto done;
	}

	source_node = adapter_get_source_node(bus_type, bus_addr, subtype);
	if (NULL == source_node)
		goto done;

	class = adapter_pci_get_field_value(source_node, "class");
	if ((NULL == class) || (strlen(class) != 4))
		goto done;

	/* It's a hack, but adapter_pci_lookup can be (ab)used like this. */
	v[0] = 'C'; v[1] = ' ';
	v[2] = class[0]; v[3] = class[1];
	v[4] = '\0';

	t = lsvpd_relative_filename(lsvpd_lib_dir, ADAPTER_PCI_IDS_FILENAME);
	if (NULL == t)
		goto done;

	if (adapter_pci_lookup(t,
			       v, &class[2], NULL, NULL,
			       &c, &ret, &x1, &x2)) {
		goto done;
		
	}

 done:
	free(source_node);
	free(class);
	free(t);
	return ret;
}

/********************************************************************/

adapter_pci_adapter_get_vpd_t * adapter_pci_retrieve_vpd_first = NULL;

vpd_list_t *
adapter_pci_retrieve_vpd(const bus_addr_t bus_addr,
			 const char * node)
{
	vpd_list_t * ret;
	node_t subtype;

	ret = NULL;

	subtype = adapter_get_subtype(ADAPTER_TYPE_PCI, bus_addr);
	if (NULL != subtype) {
		if (lsvpd_file_relative_write_string(node, LSVPD_SUBTYPE_FILE,
						      subtype)) {

			if (NULL != adapter_pci_retrieve_vpd_first) 
				ret = (*adapter_pci_retrieve_vpd_first)
					(bus_addr, subtype, node);
			
			if (NULL == ret)
				ret = pci_retrieve_pci22_vpd(bus_addr, subtype,
							     node);

			/* FIXME: implement PCI 2.0/2.1 VPD retrieval
			 * via the sysfs ROM attribute.
			 * Hmmm... that'll mean another function
			 * pointer, since this module doesn't know
			 * about sysfs. */

			if (NULL == ret)
				ret = pci_manufacture_vpd(bus_addr, subtype,
							  node);
		}

		free(subtype);
	}
	return ret;
}

/**/

/* FIXME: This is temporary.  This implementation doesn't get all the
 * VPD, just the core fields.  Later this function will be moved or
 * will be changed to call (the equivalent of) adapter_extra_vpd_hook
 * and adapter_retrieve_logicals_vpd. This may result in the above
 * function becoming static. */

static bool
adapter_pci_retrieve_vpd_partial(const node_t type,
				 const bus_info_t bus_info,
				 const char * node,
				 vpd_list_t ** vpd)
{
	/* FIXME: I can't seem to get node_find_address to return a
	 * const bus_info_t.  Hence the horrible cast further down. */
	const char * bus_addr =
		node_find_address(ADAPTER_TYPE_PCI, bus_info);

	*vpd = NULL;

	if (NULL != bus_addr)
		*vpd = adapter_pci_retrieve_vpd((const bus_info_t) bus_addr,
						node);

	return NULL != *vpd;
}

/********************************************************************/

static void
init(void)
{
	node_functions_t p = 
		{
			.type         = ADAPTER_TYPE_PCI,
			.supertype    = ADAPTER_TYPE_DEFAULT,
			.retrieve_vpd = adapter_pci_retrieve_vpd_partial,
		};
	node_type_register(&p);

	/* Remember old function to call. */
	pci_get_ds_basic  = adapter_get_ds_function;
	adapter_get_ds_function = pci_get_ds_function;
}

INIT(init);
