/*
  adapter_pci_legacy - Print PCI (and other) information for adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter_pci_legacy.c,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/

#define _GNU_SOURCE

#include <stdbool.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <linux/pci_ids.h>

#include "adapter_pci_legacy_proc.h"
#include "adapter_pci_legacy.h"
#include "adapter_pci_legacy_ide.h"
#include "adapter_pci_legacy_scsi.h"
#include "adapter_pci_legacy_ethernet.h"

/********************************************************************/

#define ADAPTER_PCI_NODOM 0xffffffff

void
adapter_pci_info_free(adapter_pci_info_t * l)
{
	adapter_pci_info_t * t;
	while (NULL != l) {
		t = l;
		l = l->next;
		if (NULL != t)
			free(t->name);
		free(t);
	}
}

static void
adapter_pci_print_info(const adapter_pci_info_t * i,
		       uint32_t dom)
{
	switch (i->class) {
	case PCI_CLASS_NETWORK_ETHERNET: printf("ethernet");  break;
	case PCI_CLASS_STORAGE_SCSI:     printf("scsi");      break;
	case PCI_CLASS_STORAGE_IDE:      printf("ide");       break;
	default:                         return;
	}

	printf(" %04x:%02hhx:%02hhx.%hhx %s",
	       (unsigned int)  dom,
	       (unsigned char) i->bus,
	       (unsigned char) i->dev,
	       (unsigned char) i->fun,
	       i->name);

	if (PCI_CLASS_NETWORK_ETHERNET == i->class) {
		int j;
		printf(" ");
		for (j = 0; j < PCI_ETHERNET_HWADDR_LEN ; j++) {
			if (j > 0)
				printf(":");
			printf("%02hhx", i->eth_hwaddr[j]);
		}
	}
	printf("\n");
}

static bool
adapter_pci_possible_match(adapter_pci_info_t * i,
			   proc_pci_list_t * p)
{
	memlist_t * m;

	/* FIXME: Gross hack for i->channel: open firmware doesn't
	 * seem to distinguish the channels anyway. */
	if ((i->class == p->class)                         &&
	    (i->bus == (p->bus & 0xff))                    &&
	    (i->dev == p->dev)                             &&
	    (i->fun == p->fun)                             &&
	    ((0 == i->irq) || ((p->irq & 0xff) == i->irq)) &&
	    (0 == i->ide_channel)) {
		if (NULL == p->mem)
			return true;
		else {
			m = p->mem;
			while (NULL != m) {
				if ((0 == i->eth_mem_start) ||
				    ((i->eth_mem_start & 0xffffffff) ==
				     (m->start & 0xffffffff)))
					return true;
				m = m->next;
			}
		}
	}
	return false;
}

static bool
adapter_pci_unique_match(proc_pci_list_t * p,
			 adapter_pci_info_t * il)
{
	adapter_pci_info_t * t;
	unsigned int num_matches = 0;

	for (t = il; NULL != t; t = t->next)
		if (adapter_pci_possible_match(t, p)) {
			num_matches++;
			if (1 < num_matches)
				return false;
		}

	return (1 == num_matches);
}

static bool
adapter_unique_pci_match(adapter_pci_info_t * i,
			 proc_pci_list_t * pl)
{
	proc_pci_list_t * t;
	unsigned int num_matches = 0;

	for (t = pl; NULL != t; t = t->next)
		if (adapter_pci_possible_match(i, t)) {
			num_matches++;
			if (1 < num_matches)
				return false;
		}

	return (1 == num_matches);
}

static void
adapter_pci_print_dodgy_matches(adapter_pci_info_t * i,
				adapter_pci_info_t * il,
				proc_pci_list_t * pl)
{
	adapter_pci_info_t * it;
	proc_pci_list_t * t;
	uint32_t dom;

	for (t = pl; NULL != t; t = t->next)
		if (adapter_pci_possible_match(i, t)) {
			dom = (uint32_t) (t->bus >> 8) & 0xffffff;
			if (adapter_pci_unique_match(t, il))
				/* PCI address matches one IDE host. */
				adapter_pci_print_info(i, dom);
			else
				/* Find a different IDE host that
				   matches the address we're looking
				   at.  If this other IDE host
				   matches only that address, then the
				   address can't belong to the current
				   adapter - otherwise print it.  */
				for (it = il; NULL != it; it = it->next)
					if ((0 != strcmp(i->name, it->name)) &&
					    adapter_pci_possible_match(it,
									   t) &&
					    (!adapter_unique_pci_match(it, pl)))
						adapter_pci_print_info(i, dom);
		}
}

/********************************************************************/

static void
adapter_pci_map(adapter_pci_gatherer_t * f,
		proc_pci_list_t * pl)
{
	adapter_pci_info_t  * il;
	adapter_pci_info_t  * t;

	if (NULL == (il = f()))
		return;

	for (t = il; NULL != t; t = t->next) {
		if (ADAPTER_PCI_NODOM != t->dom)
			adapter_pci_print_info(t, t->dom);
		else {
			adapter_pci_print_dodgy_matches(t, il, pl);
		}
	}

	adapter_pci_info_free(il);
}

/********************************************************************/

int
main(int argc, char ** argv)
{
	proc_pci_list_t * pl = proc_pci();

	adapter_pci_map(&adapter_pci_legacy_ide,      pl);
	adapter_pci_map(&adapter_pci_legacy_scsi,     pl);
	adapter_pci_map(&adapter_pci_legacy_ethernet, pl);

	proc_pci_list_free(pl);

        return 0;
}
