/*
  adapter_pci_legacy_proc - Extract useful information from /proc/pci.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter_pci_legacy_proc.c,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/
  
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "adapter_pci_legacy_proc.h"

#define PROC_PCI "/proc/pci"

static proc_pci_list_t *
proc_pci_list_new(uint32_t bus, uint8_t dev, uint8_t fun,
		  proc_pci_list_t * next)
{
	proc_pci_list_t * t =  malloc(sizeof(proc_pci_list_t));

	t->bus = bus;
	t->dev = dev;
	t->fun = fun;
		
	t->class = 0xffff;
	t->irq   = 0xffff;
	t->mem   = NULL;
		
	t->next = next;

	return t;
}

void
proc_pci_list_free(proc_pci_list_t * l)
{
	proc_pci_list_t * t;
	memlist_t * t2;

	while (NULL != l) {
		t = l;
		l = l->next;
		while (t->mem) {
			t2 = t->mem;
			t->mem = t->mem->next;
			free(t2);
		}
		free(t);
	}
}

void
proc_pci_list_fprint(FILE * f, proc_pci_list_t * l)
{
	memlist_t * t;
	unsigned long long int start;
	unsigned long long int end;

	while (NULL != l) {
		fprintf(f, "%08x:%02hhx:%02hhx %04hx %04hx",
			l->bus, l->dev, l->fun, l->class, l->irq);
		t = l->mem;
		while (NULL != t) {
			start = (unsigned long long int) t->start;
			end   = (unsigned long long int) t->end;
			fprintf(f, " 0x%08llx-0x%08llx", start, end);
			t = t->next;
		}
		fprintf(f, "\n");
		l = l->next;
	}
}

static void
proc_pci_other (const char * this_line, proc_pci_list_t * l)
{
	unsigned short int irq;
	unsigned long long int start;
	unsigned long long int end;
	memlist_t * t;

	if (adapter_pci_legacy_class(this_line, l))
		return;
	else if (2 == sscanf(this_line,
			     " %*s %*d bit memory at 0x%llx [0x%llx].",
			     &start, &end)) {
		t = malloc(sizeof(memlist_t));
		t->start = start;
		t->end   = end;
		t->next  = l->mem;
		l->mem   = t;
	} else if (1 == sscanf(this_line, " IRQ %hu.", &irq))
		l->irq = (uint16_t) irq;

	return;
}

static bool
proc_pci_line_scanner(const char * line,
		      void * x)
{
	unsigned int bus;
	unsigned int dev;
	unsigned int fun;
	proc_pci_list_t ** l;

	l = x;

	if (3 == sscanf(line,
			" Bus %u, device %u, function %u: ",
			&bus, &dev, &fun)) {
		*l = proc_pci_list_new(bus, dev, fun, *l);
	} else if (NULL != *l)
		proc_pci_other(line, *l);

	return true;
}

/* Extra level of functions to allow testing against fixed files. */
static proc_pci_list_t *
proc_pci_file(const char * filename)
{
	proc_pci_list_t * l = NULL;

	line_scanner_harness(filename, proc_pci_line_scanner, &l);
	return l;
}

proc_pci_list_t *
proc_pci()
{
	return proc_pci_file(PROC_PCI);
}


#if 0
int
main(int argc, char ** argv)
{
	proc_pci_list_t * l = proc_pci();

	if (NULL != l) {
		proc_pci_list_fprint(stderr, l);
		proc_pci_list_free(l);
		return 0;
	} else {
		return 1;
	}
}

#endif
