/*
  adapter_pci_legacy_scsi - Get PCI (and other) information for scsi adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter_pci_legacy_scsi.c,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/
  
#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <linux/pci_ids.h>

#include "util.h"
#include "adapter_pci_legacy.h"
#include "adapter_pci_legacy_scsi.h"

#define PCI_SCSI_ADAPTER_TEMPLATES LSVPD_SUBDIR "/scsihost.conf"
#define PCI_SCSI_PROC_DIR          "/proc/scsi"

static void
pci_scsi_parser_common(uint32_t      hostnum,
		       uint32_t      dom,
		       uint8_t       bus,
		       uint8_t       dev,
		       uint8_t       fun,
		       uint16_t      irq,
		       const char *  driver,
		       adapter_pci_info_t ** ps)
{
	adapter_pci_info_t * p;

	p = calloc(1, sizeof(adapter_pci_info_t));

	if (NULL != p) {
		if (-1 == asprintf(&p->name, "host%u", hostnum))
			free(p);
		else {
			p->class = PCI_CLASS_STORAGE_SCSI;

			p->dom = dom;
			p->bus = bus;
			p->dev = dev;
			p->fun = fun;
			p->irq = irq;

			p->scsi_driver = driver;

			p->ide_channel   = 0;
			p->eth_mem_start = 0;

			p->next = *ps;
			*ps = p;
		}
	}
}

typedef void (*pci_scsi_parser) (uint32_t      hostnum,
				 const char *  driver,
				 const char *  data,
				 adapter_pci_info_t ** ps);

static void
pci_scsi_sym538xx_parser (uint32_t      hostnum,
			  const char *  driver,
			  const char *  data,
			  adapter_pci_info_t ** ps)
{
	unsigned int  bus;
	unsigned char dev;
	unsigned char fun;
	unsigned int  irq;

	char * t;

	if (NULL == (t = strstr(data, "PCI bus")))
		return;
	if (4 != sscanf(t, "PCI bus %u, device %hhu, function %hhu, IRQ %u",
			&bus, &dev, &fun, &irq))
		return;

	pci_scsi_parser_common(hostnum,
			       (bus > 0xff ? bus >> 8 : ADAPTER_PCI_NODOM),
			       (uint8_t) (bus > 0xff ? bus & 0xff : bus),
			       dev,
			       fun,
			       irq,
			       driver,
			       ps);
}

static void
pci_scsi_ipr_parser (uint32_t      hostnum,
		     const char *  driver,
		     const char *  data,
		     adapter_pci_info_t ** ps)
{
	unsigned long long  hostaddr;

	char * t;

	if (NULL == (t = strstr(data, "Host Address:")))
		return;
	if (1 != sscanf(t, "Host Address: %llx", &hostaddr))
		return;

	pci_scsi_parser_common(hostnum,
			       /* Doesn't appear to have leading 0s,
				* so we just don't know! :-( */
			       (uint32_t) (((hostaddr >> 24) & 0xffffff) ||
					   ADAPTER_PCI_NODOM),
			       (uint8_t) ((hostaddr >> 16) & 0xff),
			       (uint8_t) ((hostaddr >> 8) & 0xff),
			       (uint8_t) (hostaddr & 0xff),
			       0,
			       driver,
			       ps);
}

static struct pci_scsi_driver_info {
	char * name;
	pci_scsi_parser parser;
} pci_scsi_drivers[] = {
	{ "sym53c8xx", pci_scsi_sym538xx_parser },
	{ "ipr",       pci_scsi_ipr_parser },
	{ "ibmsis",    pci_scsi_ipr_parser },
	{ NULL,        NULL }
};

/* This is horrible!  If any of the entries for a particular driver
 * have a known PCI domain, then we assume the driver prints long bus
 * numbers, so we set any unknown domains to 0.  This can save the
 * caller a lot of trouble. */
static void
hack_doms(const char * driver, adapter_pci_info_t * l)
{
	adapter_pci_info_t * t;
	bool hackdoms = false;

	for (t = l; NULL != t; t = t->next)
		if ((0 == strcmp(t->scsi_driver, driver)) &&
		    (ADAPTER_PCI_NODOM != t->dom)) {
			hackdoms = true;
			break;
		}
	if (hackdoms)
		for (t = l; NULL != t; t = t->next)
			if ((0 == strcmp(t->scsi_driver, driver)) &&
			    (ADAPTER_PCI_NODOM == t->dom))
				t->dom = 0;
}

/* Extra level of functions to allow testing against fixed files. */
static adapter_pci_info_t *
pci_scsi_file(const char * proc_scsi_dir)
{
	struct pci_scsi_driver_info * p = pci_scsi_drivers;
	char * dname = NULL;
	DIR * dir;
	struct dirent * e;
	unsigned int hostnum;
	char x;
	char * f = NULL;
	char * data = NULL;
	
	adapter_pci_info_t * ret = NULL;

	p = pci_scsi_drivers;

	for (p = pci_scsi_drivers; NULL != p->name; p++) {
		/*fprintf(stderr, "%s\n", p->name);*/

		/* dname = /proc/scsi/<driver>. */
		asprintf(&dname, "%s/%s", proc_scsi_dir, p->name);

		dir = opendir(dname);
		if (NULL == dir) {
			free(dname);
			continue;
		}

		while (NULL != (e = readdir(dir))) {
			if (1 != sscanf(e->d_name, "%u%c", &hostnum, &x))
				continue;
			asprintf(&f, "%s/%u", dname, hostnum);
			data = lsvpd_file_read_string(f);
			if (NULL != data) {
				p->parser((uint32_t) hostnum, p->name,
					  data, &ret);
				free(data);
			}
			free(f);
		}

		free(dname);
		closedir(dir);
	}

	for (p = pci_scsi_drivers; NULL != p->name; p++)
		hack_doms(p->name, ret);

	return ret;
}


adapter_pci_info_t *
adapter_pci_legacy_scsi(void)
{
	return pci_scsi_file(PCI_SCSI_PROC_DIR);
}
