/*
  adapter_pci_lookup - Lookup PCI vendors and devices in pci.ids
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2003, 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter_pci_lookup.c,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/
  
/* For asprintf: */
#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "util.h"

#include "adapter_pci_lookup.h"

static size_t
length_or_0(const char * s)
{
	return (NULL != s) ? strlen(s) : 0;
}

/* Check haystack for needle at offset. Ensure it is followed by 2
 * spaces.  If so, duplicate the rest of haystack. */
static bool
check_and_dup(const char * haystack,
	      size_t offset,
	      const char * needle,
	      size_t len,
	      char ** out)
{
	if ((0 != len) &&
	    (0 == strncasecmp(haystack+offset, needle, len)) &&
	    (' ' == haystack[offset+len]) &&
	    (' ' == haystack[offset+len+1])) {
		*out = strdup(&haystack[offset+len+2]);
		return true;
	} else
		return false;
}

struct adapter_pci_lookup_info {
	const char * vendor_id;
	const char * device_id;
	const char * sub_ids;
	size_t vendor_id_len;
	size_t device_id_len;
	size_t sub_ids_len;
	size_t subvendor_len;
	bool got_vendor;
	bool got_device;
	char * vendor_string;
	char * device_string;
	char * subvendor_string;
	char * subdevice_string;
};

static bool
adapter_pci_lookup_line_scanner(const char * line,
				void * x)
{
	struct adapter_pci_lookup_info * i;

	i = x;

	/* Comment: next line please! */
	if ('#' == line[0])
		return true;

	/* If there's no leading TAB, we're possibly looking at a
	 * vendor id. */
	if ('\t' != line[0]) {
		/* If we already have our vendor strings then we're done. */
		if ((NULL != i->vendor_string) &&
		    ((NULL == i->sub_ids) ||
		     ((NULL != i->sub_ids) &&
		      (NULL != i->subvendor_string))))
			return false;

		i->got_device = false;
		
		/* No leading tab, might be vendor line. */
		i->got_vendor = check_and_dup(line, 0,
					      i->vendor_id, i->vendor_id_len,
					      &i->vendor_string);

		/* This isn't an "else" - vendor_id and subvendor_id
		 * may be the same, although the latter doesn't affect
		 * i->got_vendor.. */
		check_and_dup(line, 0, i->sub_ids, i->subvendor_len,
			      &i->subvendor_string);
	} else if (i->got_vendor) {
		/* We're in the required vendor section and we've got
		   a leading TAB.  Look for device id and subdevice id. */
		if ('\t' != line[1])
			/* Just 1 leading tab, might be device line. */
			i->got_device = check_and_dup(line, 1,
						      i->device_id,
						      i->device_id_len,
						      &i->device_string);
		else if (i->got_device)
			check_and_dup(line, 2, i->sub_ids, i->sub_ids_len,
				      &i->subdevice_string);
	}

	return true;
}

bool
adapter_pci_lookup(const char * filename,
		   const char * vendor_id,
		   const char * device_id,
		   const char * subvendor_id,
		   const char * subdevice_id,
		   char ** vendor_string,
		   char ** device_string,
		   char ** subvendor_string,
		   char ** subdevice_string)
{
	bool ret;
	struct adapter_pci_lookup_info i;
	char * t;

	i.vendor_id = vendor_id;
	i.device_id = device_id;

	/* Create combined sub_ids field. */
	if (NULL != subvendor_id) {
		if (NULL == subdevice_id)
			t = strdup(subvendor_id);
		else if (-1 == asprintf(&t,
					"%s %s", subvendor_id, subdevice_id))
			t = NULL;
	} else
		t = NULL;
	i.sub_ids = t;

	/* Precalculate  length  of   all  input  strings.   Premature
	 * optimisation! */
	i.vendor_id_len = length_or_0(i.vendor_id);
	i.device_id_len = length_or_0(i.device_id);
	i.sub_ids_len   = length_or_0(i.sub_ids);
	if (NULL != i.sub_ids)
		i.subvendor_len = length_or_0(subvendor_id);
	else
		i.subvendor_len = 0;

	i.got_vendor = false;
	i.got_device = false;

	i.vendor_string = NULL;
	i.device_string = NULL;
	i.subvendor_string = NULL;
	i.subdevice_string = NULL;

	ret = line_scanner_harness(filename,
				   adapter_pci_lookup_line_scanner, &i);

	free(t);

	*vendor_string = i.vendor_string;
	*device_string = i.device_string;
	*subvendor_string = i.subvendor_string;
	*subdevice_string = i.subdevice_string;

	return ret;
}
