/*
  device - VPD support routines for devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_scsi.c,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "lsvpd.h"
#include "init.h"
#include "util.h"

#include "device.h"

#include "device_scsi.h"
#include "device_ide.h"

/********************************************************************/

bool
device_parse_scsi_address(const char * s,
			  device_scsi_address_t * a)
{
	return (4 == sscanf(s,
			    "%u:%u:%u:%u",
			    &a->host, &a->bus, &a->target, &a->lun));
}

bool
device_parse_scsi_bus_info(const bus_info_t bus_info,
			   device_scsi_address_t * a)
{
	const char * ba;

	ba = node_find_address(DEVICE_TYPE_SCSI, bus_info);
	return ((NULL != ba) && device_parse_scsi_address(ba, a));
}

/********************************************************************/

/* Very vaguely taken from linux-2.5.22/drivers/scsi/scsi.[ch]
 * There are only really 14 possible type, but filling it out to 16
 * removes the need for error checking. */
static const struct string_array device_scsi_types_short[] =
{
        {0x0, "disk"},
        {0x1, "tape"},
        {0x2, "printer"},
        {0x3, "processor"},
        {0x4, "worm"},
        {0x5, "cdrom"},
        {0x6, "scanner"},
        {0x7, "optical-disk"},
        {0x8, "jukebox"},
        {0x9, "communications"},
        {0xa, "graphics"},
        {0xb, "graphics"},
        {0xd, "enclosure"},
        {-1, NULL}
};

static const struct string_array device_scsi_types[] =
{
        {0x0, "Disk Drive"},
        {0x1, "Tape Drive"},
        {0x2, "Printer Device"},
        {0x3, "Processor Device"},
        {0x4, "Write-once Device"},
        {0x5, "CD-ROM Drive"},
        {0x6, "Scanner Device"},
        {0x7, "Optical Memory Device"},
        {0x8, "Medium Changer Device"},
        {0x9, "Communications Device"},
        {0xa, "Graphics Arts Pre-press Device"},
        {0xb, "Graphics Arts Pre-press Device"},
        {0xd, "Enclosure Services Device"},
        {-1, NULL}
};


const char *
device_scsi_subtype_string(unsigned char subtype)
{
	return string_array_lookup(device_scsi_types_short,
				   "unknown",
				   subtype);
}

/* Note: SATA could say "Serial ATA".  However, a SATA controller
 * might have non-SATA devices attached to legacy ports, so we just
 * say "ATA".
 */
struct device_scsi_ds_foo device_scsi_ds_prefixes[] = {
	{ DEVICE_TYPE_SCSI,       "SCSI"            },
	{ DEVICE_TYPE_SATA,       "ATA"             },
	{ DEVICE_TYPE_ATA,        "ATA"             },
	{ DEVICE_TYPE_ISCSI,      "iSCSI"           },
	{ DEVICE_TYPE_FIBRE,      "Fibre Channel"   },
	{ DEVICE_TYPE_RAID,       "SCSI"            },
	{ DEVICE_TYPE_SCSI_DEBUG, "SCSI Debug Fake" },
	{ DEVICE_TYPE_IBMVSCSI,   "Virtual SCSI"    },
	{ DEVICE_TYPE_USB,        "USB"             },
	{ NULL,                   NULL              },
};

char *
device_scsi_ds(const node_t type,
	       unsigned char  subtype)
{
	char * ret;
	int i;
	node_t x;
	const char * p;
	const char * s;

	ret = NULL;

	p = NULL;
	if (NULL != type) {
		i = 0;
		while ((NULL != (x = device_scsi_ds_prefixes[i].type)) &&
		       (0 != strcmp(x, type)))
			i++;
		p = device_scsi_ds_prefixes[i].ds_prefix;
	}
	if (NULL == p)
		p = "SCSI";

	/* string_array_lookup can not fail. */
	s = string_array_lookup(device_scsi_types, "Unknown Device", subtype);

	/* When we get 2 special cases we will generalise this using a
	 * sparse lookup-table like the above string array, but keyed
	 * off both type and subtype. */
	if ((0 == strcmp("USB", p)) && (0 == strcmp("Disk Drive", s)))
		s = "Mass Storage Device";

	if (-1 == asprintf(&ret, "%s %s", p, s ))
		ret = NULL;

	return ret;
}

void
device_scsi_ata_mf_tm(const node_t type,
		      char ** mf,
		      char ** tm)
{
	char * t;
	char * x;

	if (0 == strcmp(*mf, "ATA")) {
		/* For drivers that set MF to "ATA", rebuild MF from
		 * TM, possibly hacking TM. */
		free(*mf);
		device_ide_do_mf_tm(mf, tm);
	} else if ((0 == strcmp(type, DEVICE_TYPE_SATA)) ||
		   (0 == strcmp(type, DEVICE_TYPE_ATA))) {
		/* This is a bit gross and hacky, but it does the job. */
		x = strdup(*mf);
		if (NULL != x) {
			if (device_ide_do_mf_tm(&t, mf)) {
				free(x);
				free(*mf);
				*mf = t;
			} else {
				free(t);
				free(*mf);
				*mf = x;
			}
		}
	}
}

/********************************************************************/

bool
device_scsi_append_field(vpd_list_t **  vpd,
			 const char *   key,
			 const unsigned char * data,
			 size_t         len,
			 const char *   subtype_string)
{
	bool ret;
	char * ldata;

	ret = false;

	/* Certain fields will require some munging. */
	/* Convert RL, Z0, Z7 to hex. */
	if (((strcmp(key, "RL") == 0) &&
	     (strcmp(subtype_string, "disk") == 0)) ||
	    (strcmp(key, "Z0") == 0) ||
	    (strcmp(key, "Z7") == 0))
		/* lsvpd_hexify returns a new NUL-terminated string. */
		ldata = lsvpd_hexify(data, len);
	else
		/* Make trimmed, NUL-terminated copy. */
		ldata = strndup_trim(data, len);
	
	if (NULL != ldata) {
		/* Write data to file. */
		ret = vpd_list_maybe_append_key_string(vpd, key,
						       ldata);
		free(ldata);
	}

	return ret;
}

bool
device_scsi_append_field_string(vpd_list_t **  vpd,
				const char *   key,
				const char *   data,
				const char *   subtype_string)
{
	return (NULL != data) &&
		device_scsi_append_field(vpd, key,
					 (unsigned char *) data, strlen(data),
					 subtype_string);
}

/********************************************************************/

static bool
device_scsi_validate (const node_t     type,
		      const bus_info_t bus_info)
{
	bool ret;
	device_scsi_address_t a;

	ret = true;

	if (device_parse_scsi_bus_info(bus_info, &a) &&
	    (0xff == a.bus) &&
	    (0xff == a.target) &&
	    (0xff == a.lun))
		ret = false;

	return ret;
}

static struct device_scsi_vpd_foo {
	node_retrieve_vpd_function_t * f;
	struct device_scsi_vpd_foo *   next;
} * device_scsi_vpd_functions = NULL;

void
device_scsi_vpd_function_append(node_retrieve_vpd_function_t * f)
{
	struct device_scsi_vpd_foo * t;

	if (NULL == device_scsi_vpd_functions) {
		t = malloc(sizeof(struct device_scsi_vpd_foo));
		device_scsi_vpd_functions = t;
	} else {
		t = device_scsi_vpd_functions;
		while (NULL != t->next)
			t = t->next;
		t->next = malloc(sizeof(struct device_scsi_vpd_foo));
		t = t->next;
	}
		
	if (NULL != t) {
		t->f = f;
		t->next = NULL;
	}
}


static bool
device_scsi_retrieve_vpd(const node_t type,
			 const bus_info_t bus_info,
			 const char * node,
			 vpd_list_t ** vpd)
{
	bool ret;
	struct device_scsi_vpd_foo * t;
	
	ret = false;

	if (device_scsi_validate(type, bus_info)) {
		t = device_scsi_vpd_functions;
		while ((!ret) && (NULL != t)) {
			ret = t->f(type, bus_info, node, vpd);
			t = t->next;
		}
	}

	return ret;
}

static char *
device_scsi_get_parent_name (const node_t type,
			     const bus_info_t bus_info)
{
	char * ret;
	device_scsi_address_t a;

	ret = NULL;

	if (device_parse_scsi_bus_info(bus_info, &a)) {
		// FIXME: "host" should not be hardcoded.
		if (-1 == asprintf(&ret, "host%d", a.host))
			ret = NULL;
	}

	return ret;
}

static const char * device_scsi_vpd_sort_order[] =
{
	"FC", "DS", "AX", "PL", "MF", "TM", "CD", "LL",
	"YL", "FN", "RL", "SN", "EC", "PN",
	NULL
};

void
device_scsi_vpd_sort(vpd_list_t ** vpd)
{
	vpd_list_sort(vpd, device_scsi_vpd_sort_order);
}

device_get_info_function_t * device_scsi_get_info_function = NULL;

device_node_info_t *
device_scsi_get_info (const node_t     type,
		      const bus_info_t bus_info)
{
	return (NULL != device_scsi_get_info_function) ?
		(device_scsi_get_info_function)(type, bus_info):
		NULL;
}

/********************************************************************/

/*
 * Defaults:
 *   retrieve_vpd returns false.
 */

static void
init(void)
{
	int i;
	node_t x;
	node_functions_t s = 
		{
			.type             = DEVICE_TYPE_SCSI,
			.supertype        = DEVICE_TYPE_DEFAULT,
			.retrieve_vpd     = device_scsi_retrieve_vpd,
			.get_parent_name  = device_scsi_get_parent_name,
		};
	node_type_register(&s);

	/* Register other types, with appropriate supertype. */
	s.supertype = DEVICE_TYPE_SCSI;
	i = 0;
	while (NULL != (x = device_scsi_ds_prefixes[i].type)) {
		if (DEVICE_TYPE_SCSI != x) {
			s.type = x;
			node_type_register(&s);
		}
		i++;
	}
}

INIT(init);
