/*
  device - sysfs support routines for devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_sysfs.c,v 1.3 2006/09/11 18:40:25 munsone Exp $
*/

/* For asprintf(3). */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <glob.h>

#include "util.h"
#include "vpd.h"

#include "lsvpd.h"
#include "init.h"
#include "sysfs.h"
#include "libsysfs.h"

#include "device.h"
#include "device_scsi.h"
#include "device_ide.h"

/********************************************************************/

static struct {
	node_t type;
	char *   sysfs_bus;
} device_sysfs_type_alist[] = {
	{ DEVICE_TYPE_IDE,     "ide"   },
	{ DEVICE_TYPE_SCSI,    "scsi"  },
	{ DEVICE_TYPE_DEFAULT, NULL },
};

static const char *
device_sysfs_type_lookup(const node_t type)
{
	const char * ret;
	int i;

	ret = NULL;

	i = 0;
	while ((NULL == ret) &&
	       (NULL != device_sysfs_type_alist[i].sysfs_bus)) {
		if (0 == strcmp(type, device_sysfs_type_alist[i].type)) {
			ret = device_sysfs_type_alist[i].sysfs_bus;
		}
		i++;
	}

	return ret;
}

/********************************************************************/

static bool
device_sysfs_list_devices(bus_info_list_t ** node_list)
{
	bool ret;

	int i;
	node_t t;
	char * s;
	struct dlist * names;
	struct sysfs_bus* bus;
	char * c;
	bus_info_t bi;

	ret = true;

	node_list_clear(node_list);

	i = 0;
	while (NULL != (s = device_sysfs_type_alist[i].sysfs_bus)) {
		t = device_sysfs_type_alist[i].type;
		bus = sysfs_open_bus( s );
		if (NULL != bus) {
			names = sysfs_get_bus_devices(bus);
			if (NULL != names) {
				dlist_for_each_data(names, c, char) {
					if (-1 != asprintf(&bi, "%s%c%s", t,
						   	BUS_INFO_SEP, c)) {
						ret = ret &&
							node_list_add(node_list, t, bi);
						free(bi);
					}
				}
			}
			sysfs_close_bus(bus);
		}
		i++;
	}

	return ret;
}

/********************************************************************/

static struct {
	const char * pattern;
	mode_t mode;
} device_sysfs_links[] = {
	{ "block:*",        S_IFBLK },
	{ "block",          S_IFBLK },
	{ "scsi_tape:*",    S_IFCHR },
	{ "tape",           S_IFCHR },
	{ "scsi_generic:*", S_IFCHR },
	{ "generic",        S_IFCHR },
	{ NULL,             S_IFREG },
};

static bool
device_sysfs_find_link(const node_t type,
		       const bus_info_t bus_info,
		       char ** link,
		       mode_t * mode)
{
	bool ret;
	char * d;
	struct stat statbuf;
	glob_t globbuf;
	char * l;
	int i;

	ret = false;

	// FIXME: Should have something better than NULL here.
	// However, since we're in sysfs-specific code we know that
	// argument is ignored.
	d = node_get_source_dir(type, bus_info, NULL); 

	if (NULL == d)
		goto done;

	/* stat(2) here, not lstat(2) - making sure there's an
	 * underlying directory. */
	if (-1 == stat(d, &statbuf))
		goto free_d;

	if (!S_ISDIR(statbuf.st_mode))
		goto free_d;

	i = 0;
	l = NULL;
	while ((!ret) && (NULL != device_sysfs_links[i].pattern)) {
		l = lsvpd_relative_filename(d, device_sysfs_links[i].pattern);
		if (NULL != l) {
			ret = (0 == glob(l, 0, NULL, &globbuf)) &&
				(globbuf.gl_pathc > 0) &&
				(0 == lstat(globbuf.gl_pathv[0], &statbuf)) &&
				S_ISLNK(statbuf.st_mode);
			if (ret) {
				*link = strdup(globbuf.gl_pathv[0]);
				*mode = device_sysfs_links[i].mode;
			}

			globfree(&globbuf);
			free(l);
		}

		i++;
	}
					
 free_d:
	free(d);

	/* Sanity check on above strdup. */
	ret = ret && (NULL != *link);

 done:
	return ret;
}

static struct {
	const char * from;
	const char * to;
} device_sysfs_prefix_subst_alist[] = {
	{ "sr",  "scd" },
	{ NULL,  NULL  },
};

static void
device_sysfs_prefix_substitute(char ** s)
{
	int i;
	const char * f;

	i = 0;
	while ((NULL != (f = device_sysfs_prefix_subst_alist[i].from)) &&
	       !lsvpd_prefix_substitute(s,
					f,
					device_sysfs_prefix_subst_alist[i].to))
		i++;
}

static char *
device_sysfs_get_directory_attribute(const char * d,
				     const char * a)
{
	char * ret;
	char * attr_path;
	struct sysfs_attribute * sa;
	int length;

	ret = NULL;

	length = strlen( d ) + strlen( a ) + 2;
	attr_path = malloc( length );

	if( NULL == attr_path )
	{
		return ret;
	}

	memset( attr_path, '\0', length );
	strcpy( attr_path, d );
	strcat( attr_path, "/" );
	strcat( attr_path, a );

	sa = sysfs_open_attribute( attr_path );
	if( NULL != sa )
	{
		if( 0 == sysfs_read_attribute( sa ) )
		{
			ret = strndup_trim( (unsigned char*)sa->value, sa->len );
		}
	}

	free( attr_path );

	sysfs_close_attribute( sa );

	return ret;
}

/********************************************************************/

static char *
device_sysfs_get_kernel_name(const node_t type,
			     const bus_info_t bus_info)
{
	char * ret;
	char * l;
	mode_t x;
	char * t;

	ret = NULL;

	if (device_sysfs_find_link(type, bus_info, &l, &x)) {
		t = lsvpd_readlink(l);
		if (NULL != t) {
			ret = lsvpd_basename(t);
			device_sysfs_prefix_substitute(&ret);
			free(t);
		}
		free(l);
	}

	return ret;
}

static device_node_info_t *
device_sysfs_get_info(const node_t type,
		      const bus_info_t bus_info)
{
	device_node_info_t * ret;
	char * l;
	mode_t m;
	char * mm;

	ret = NULL;

	if (device_sysfs_find_link(type, bus_info, &l, &m)) {
		mm = lsvpd_file_relative_read_string(l, "dev");
		if (NULL != mm) {
			ret = malloc(sizeof(device_node_info_t));
			if (NULL != ret) {
				ret->mode = m;
				if ((S_IFREG == ret->mode) ||
				    (2 != sscanf(mm, "%u:%u",
						 &ret->major,
						 &ret->minor))) {
					free(ret);
					ret = NULL;
				}
			}
			free(mm);
		}
		free(l);
	}

	return ret;
}

/********************************************************************/

static bool
device_sysfs_retrieve_vpd_scsi(const node_t type,
			       const bus_info_t bus_info,
			       const char * node,
			       vpd_list_t ** vpd)
{
	bool ret;
	const char * s;
	const char * t;
	struct sysfs_device * d;
	char * st;
	unsigned int x;
	unsigned char subtype;
	const char * subtype_string;
	char * mf;
	char * tm;
	char * rl;
	char * ds;

	ret = false;

	s = device_sysfs_type_lookup(DEVICE_TYPE_SCSI);
	t = node_find_address(DEVICE_TYPE_SCSI, bus_info);
	if ((NULL == s) || (NULL == t))
		goto done;

	d = sysfs_open_device(s, t);
	if (NULL == d)
		goto done;

	st = device_sysfs_get_directory_attribute(d->path, "type");
	mf = device_sysfs_get_directory_attribute(d->path, "vendor");
	tm = device_sysfs_get_directory_attribute(d->path, "model");
	rl = device_sysfs_get_directory_attribute(d->path, "rev");

	if (1 != sscanf(st, "%u", &x))
		goto free_fields;
	subtype = (unsigned char) x;
	subtype_string = device_scsi_subtype_string(subtype);

	if (lsvpd_mkdir(node) &&
	    lsvpd_file_relative_write_string(node, LSVPD_SUBTYPE_FILE,
					     subtype_string)) {
		ds = device_scsi_ds(type, subtype);
		if (NULL != ds) {
			ret = true;

			if ((NULL != tm) &&
			    (NULL != mf))
				device_scsi_ata_mf_tm(type, &mf, &tm);

			device_scsi_append_field_string(vpd, "TM", tm,
							subtype_string);
			device_scsi_append_field_string(vpd, "MF", mf,
							subtype_string);
			device_scsi_append_field_string(vpd, "RL", rl,
							subtype_string);
			device_scsi_append_field_string(vpd, "DS", ds,
							subtype_string);

			free(ds);

			if (NULL != vpd)
				device_scsi_vpd_sort(vpd);
		}
	}

 free_fields:
	free(st);
	free(mf);
	free(tm);
	free(rl);

	sysfs_close_device(d);
 done:
	return ret;
}

/********************************************************************/

static void
init(void)
{
	if (NULL != lsvpd_sysfs_dir) {
		node_functions_t d = {
			.type            = DEVICE_TYPE_DEFAULT,
			.get_kernel_name = device_sysfs_get_kernel_name,
		};

		node_functions_t s = {
			.type            = DEVICE_TYPE_SCSI,
			.get_kernel_name = NODE_UNDEF,
		};

		node_functions_t i = {
			.type            = DEVICE_TYPE_IDE,
			.get_kernel_name = NODE_UNDEF,
		};

		device_ide_get_info_function  = device_sysfs_get_info;
		device_scsi_get_info_function = device_sysfs_get_info;

		node_type_register(&d);
		node_type_register(&s);
		node_type_register(&i);

		node_listing_functions_clear();
		node_listing_function_register(device_sysfs_list_devices);

		device_scsi_vpd_function_append(device_sysfs_retrieve_vpd_scsi);
	}
}

INIT(init);
