#ifndef LSVPD_INIT_H
#define LSVPD_INIT_H

/*
  init - Modules initialisation code for lsvpd.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: init.h,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/
  
/*
 * ELF section initialisation magic. Allows a function to be called
 * before main.  The idea is for such functions get called in link
 * order, so they are useful for doing run-time configuration with
 * overrides.
 */

typedef void (*initcall_t)(void);

/*
 * We want a method of overriding functions (or function pointers)
 * according to a check in each module (*.o file), however we don't
 * want to keep the logic distributed throughout the modules.  This
 * means we want each module to do some magic initialisation before
 * main() is called, but we want it to happen in a predetermined order
 * - link order sounds good.
 *
 * Function pointers in the .init_array section seem to get called in
 * link order, without needing to execute something like call_inits()
 * in main().  However, the .init_array sections seems a bit new and
 * don't seem to work with an older libc (or ld.so or gcc? the
 * documentation is virtually non-existent), so we use our own ELF
 * section and execute call_inits() at the beginning of main().
 *
 * Another proposed way would be to use the "constructor" attribute on
 * each of the initialisation functions.  However, the order they are
 * executed in is unclear.  :-(
 *
 * Thanks to Rusty Russell, Jeremy Kerr and David Gibson for hints on
 * all of this.
 *
 * Building with -funit-at-a-time will result in a link failure, since
 * it optimises away pretty much everything to do with this code.
 * This means that building with -O2 will fail, since it implies the
 * above optimisation.  However, it is possible to build with -O2 by
 * adding -fno-unit-at-a-time.  Thanks to Olaf Hering <olh@suse.de>
 * for working this out.
 */
 
 /*
  * When building for GCC 4.1.0 ( perhaps 4.0.X and/or >4.1.0) the -O
  * option will optimise all this away, so you need to include the 
  * -fno-unit-at-a-time option.
  */

#define INIT(fn) static initcall_t __initcall_##fn \
__attribute__((__unused__)) \
__attribute__((__section__("init_call"))) = &fn

static inline void call_inits (void)
{
	extern initcall_t __start_init_call[], __stop_init_call[];
	initcall_t *p;
	for (p = __start_init_call; p < __stop_init_call; p++)
		(*p)();
 }

#endif
