/*
  lsvpd_node - General definitions for the lsvpd package.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: node.c,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/sysmacros.h>

#include "util.h"
#include "vpd.h"

#include "lsvpd.h"
#include "init.h"

#include "node.h"

/********************************************************************/

const char *
node_find_address(const node_t     type,
		  const bus_info_t bus_info)
{
	const char * ret;
	size_t ldst;
	size_t lbi;

	ret = NULL;

	ldst = strlen(type);
	lbi = strlen(bus_info);

	if ((lbi > ldst) &&
	    (0 == memcmp(bus_info, type, ldst)) &&
	    (BUS_INFO_SEP == bus_info[ldst]))
		ret = bus_info + ldst + 1;

	return ret;
}

bool
bus_info_parse(bus_info_t bus_info,
	       node_t * bus_type,
	       bus_addr_t * bus_addr)
{
	bool ret;

	ret = false;

	*bus_type = strdup(bus_info);
	if (NULL != bus_type) {
		*bus_addr = strchr(*bus_type, BUS_INFO_SEP);

		if ((NULL != *bus_addr) && ('\0' != *(*bus_addr + 1))) {
			*(*bus_addr) = '\0';
			/* Otherwise only need to free one result and
			 * that would be inconsistent. */
			*bus_addr = strdup(*bus_addr + 1);
			if (NULL != *bus_addr)
				ret = true;
			else
				free(bus_type);
		} else
			free(bus_type);
	}

	return ret;
}

/********************************************************************/

node_functions_t * node_functions = NULL;

static node_functions_t *
node_find_type(const node_t type)
{
	node_functions_t * t;

	t = node_functions;

	while ((NULL != t) && (0 != strcmp(t->type, type))) {
		t = t->next;
	}

	return t;
}

#define NODE_SET_POINTER(f) \
		if (NODE_UNDEF == i->f) \
			d->f = NULL; \
		else if (NULL != i->f) \
			d->f = i->f;


/* Copies i if necessary. */
void
node_type_register(node_functions_t * i)
{
	node_functions_t * t;
	node_functions_t * d;

	t = node_find_type(i->type);

	if (NULL != t) {
		d = t;
	} else {
		d = malloc(sizeof(node_functions_t));
		memset(d, 0, sizeof(node_functions_t));
		d->type = i->type;
	}

	if (NULL != d) {
		NODE_SET_POINTER(supertype);
		NODE_SET_POINTER(retrieve_vpd);
		NODE_SET_POINTER(get_parent_name);
		NODE_SET_POINTER(get_kernel_name);
		NODE_SET_POINTER(get_name);
		//NODE_SET_POINTER(set_yl);
		//NODE_SET_POINTER(add_hook);
		NODE_SET_POINTER(get_source_dir);
		NODE_SET_POINTER(get_parent_subtype);

		if (NULL == t) {
			d->next = node_functions;
			node_functions = d;
		}
	}
}

#define NODE_MULTIPLEXED_VOID(name, argdecls, args) \
void \
node_ ## name argdecls \
{ \
	node_functions_t * t; \
	void (*f) argdecls; \
\
	t = node_find_type(type); \
	f = NULL; \
	while ((NULL == f) && (NULL != t)) { \
		f = t->name; \
		if (NULL == f) \
			t = (NULL != t->supertype) ? \
				node_find_type(t->supertype) : \
				NULL; \
	} \
\
	if (NULL != f) \
		(*f) args; \
}

#define NODE_MULTIPLEXED_RET(name, argdecls, args, ret_t, ret_def) \
ret_t \
node_ ## name argdecls \
{ \
	ret_t ret; \
	node_functions_t * t; \
	ret_t (*f) argdecls; \
\
	ret = ret_def; \
\
	t = node_find_type(type); \
	f = NULL; \
	while ((NULL == f) && (NULL != t)) { \
		f = t->name; \
		if (NULL == f) \
			t = (NULL != t->supertype) ? \
				node_find_type(t->supertype) : \
				NULL; \
	} \
\
	if (NULL != f) \
		ret = (*f) args; \
\
	return ret; \
}

NODE_MULTIPLEXED_RET(retrieve_vpd,
		     (const node_t type,
		      const bus_info_t bus_info,
		      const char * node,
		      vpd_list_t ** vpd),
		     (type, bus_info, node, vpd),
		     bool,
		     false)

NODE_MULTIPLEXED_RET(get_parent_name,
		     (const node_t type,
		      const bus_info_t bus_info),
		     (type, bus_info),
		     char *,
		     NULL)

NODE_MULTIPLEXED_RET(get_name,
		     (const node_t type,
		      const bus_info_t bus_info),
		     (type, bus_info),
		     char *,
		     NULL)


NODE_MULTIPLEXED_RET(get_kernel_name,
		     (const node_t type,
		      const bus_info_t bus_info),
		     (type, bus_info),
		     char *,
		     NULL)

NODE_MULTIPLEXED_RET(get_source_dir,
		     (const node_t type,
		      const bus_addr_t bus_info,
		      const node_t subtype),
		     (type, bus_info, subtype),
		     char *,
		     NULL)

NODE_MULTIPLEXED_RET(get_parent_subtype,
		     (const node_t type,
		      const bus_addr_t bus_addr),
		     (type, bus_addr),
		     node_t,
		     NULL)

/********************************************************************/

/* Device listing. */

node_listing_function_list_t * node_listing_functions = NULL;

void node_listing_functions_clear()
{
	node_listing_function_list_t * t1;
	node_listing_function_list_t * t2;

	t1 = node_listing_functions;
	while (NULL != t1) {
		t2 = t1->next;
		free(t1);
		t1 = t2;
	}
	node_listing_functions = NULL;
}

void
node_listing_function_register(node_listing_function_t * f)
{
	node_listing_function_list_t * t;

	t = malloc(sizeof(node_listing_function_list_t));
	if (NULL != t) {
		t->f = f;
		t->next = node_listing_functions;
		node_listing_functions = t;
	}
}

bool
node_list_devices(bus_info_list_t ** node_list)
{
	bool ret;
	node_listing_function_list_t * t;

	ret = true;

	*node_list = NULL;
	for (t = node_listing_functions; NULL != t; t = t->next)
		ret = ret && t->f(node_list);

	return ret;
}

bool
node_list_add(bus_info_list_t ** node_list,
	      const node_t       type,
	      const bus_info_t   bus_info)
{
	bool ret;
	bus_info_list_t * t;

	ret = false;

	if (NULL != (t = malloc(sizeof(bus_info_list_t)))) {
		if (NULL != (t->bus_info = strdup(bus_info))) {
			t->next = *node_list;
			*node_list = t;
			ret = true;
		} else
			free(t);
	}

	return ret;
}

void
node_list_clear(bus_info_list_t ** node_list)
{
	bus_info_list_t * t1;
	bus_info_list_t * t2;

	t1 = *node_list;
	while (NULL != t1) {
		t2 = t1->next;
		free(t1->bus_info);
		free(t1);
		t1 = t2;
	}
	*node_list = NULL;
}

/********************************************************************/

static char *
node_get_name_default(const node_t     type,
		      const bus_info_t bus_info)
{
	return node_get_kernel_name(type,
				    bus_info);
}

/********************************************************************/

/* Support functions. */

bool
bus_alias_set_type(const char * bus_info,
		   const char * target,
		   const char * alias_type)
{
	bool ret;
	char * dir;
	char * k;

	ret = false;

	dir = lsvpd_relative_filename(lsvpd_db_bus_dir, bus_info);
	if (NULL != dir) {
		k = lsvpd_relative_filename(dir, alias_type);
		if (NULL != k) {
			if (0 != access(k, R_OK))
				ret = lsvpd_mkdir(dir) &&
					lsvpd_file_write_string(k, target);
			free(k);
		}
		free(dir);
	}

	return ret;
}

bool
bus_alias_set(const char * bus_info,
	      const char * target)
{
	return bus_alias_set_type(bus_info, target, LSVPD_BUS_ALIAS);
}

bool
bus_alias_clear_type(const char * bus_info,
		     const char * alias_type)
{
	bool ret;
	char * dir;
	char * k;

	ret = false;

	dir = lsvpd_relative_filename(lsvpd_db_bus_dir, bus_info);
	if (NULL != dir) {
		k = lsvpd_relative_filename(dir, alias_type);
		if (NULL != k) {
			ret = (0 == unlink(k));
			free(k);
		}
		rmdir(dir); /* Don't care about failure. */
		free(dir);
	}

	return ret;
}

bool
bus_alias_clear(const char * bus_info)
{
	return bus_alias_clear_type(bus_info, LSVPD_BUS_ALIAS);
}

char *
bus_alias_get_type(const char * bus_info,
		   const char * alias_type)
{
	char * ret;
	char * dir;

	ret = NULL;

	dir = lsvpd_relative_filename(lsvpd_db_bus_dir, bus_info);
	if (NULL != dir) {
		ret = lsvpd_file_relative_read_string(dir, alias_type);
		free(dir);
	}

	return ret;
}


char *
bus_alias_get(const char * bus_info)
{
	return bus_alias_get_type(bus_info, LSVPD_BUS_ALIAS);
}

/**/

char *
retrieve_node(const char * bus_info)
{
	return bus_alias_get(bus_info);
}

node_t
retrieve_subtype_basic(const char * node)
{
	node_t ret;

	ret = NULL;

	if (NULL != node)
		ret = lsvpd_file_relative_read_string(node, LSVPD_SUBTYPE_FILE);

	return ret;
}

node_t
retrieve_subtype(const node_t bus_type,
		 const bus_addr_t bus_addr)
{
	node_t ret;
	char * t;
	char * node;

	ret = NULL;

	t = lsvpd_relative_filename(bus_type, bus_addr);
	if (NULL != t) {
		node = retrieve_node(t);
		if (NULL != node) {
			ret = retrieve_subtype_basic(node);
			free(node);
		}
		free(t);
	}

	return ret;
}

/********************************************************************/

/*
 * No need to define dummy functions for all functions, since those
 * functions will do the appropriate thing (fail) is no functions are
 * defined in node_functions.
 */

static void
init(void)
{
	node_functions_t i = 
		{
			.type     = NODE_TYPE_DEFAULT,
			.get_name = node_get_name_default,
		};
	node_type_register(&i);
}

INIT(init);
