#ifndef LSVPD_NODE_H
#define LSVPD_NODE_H

/*
  lsvpd_node - General definitions for the lsvpd package.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: node.h,v 1.1.1.1 2013/09/12 09:32:49 bpeters Exp $
*/
  
#include <sys/types.h>
#include <stdbool.h>

#include "vpd.h"

typedef char * node_t;
typedef char * bus_info_t;
typedef char * bus_addr_t;


#define BUS_INFO_SEP '/'
#define BUS_TYPE_UNKNOWN "unknown"

extern const char *
node_find_address(const node_t type,
		  const bus_info_t bus_info);

extern bool
bus_info_parse(bus_info_t bus_info,
	       node_t * bus_type,
	       bus_addr_t * bus_addr);

/* For device listing.  This list is logically unordered. */
typedef struct bus_info_list {
	bus_info_t bus_info;
	struct bus_info_list * next;
} bus_info_list_t;

/********************************************************************/

typedef bool
(node_retrieve_vpd_function_t) (const node_t type,
				const bus_info_t bus_info,
				const char * node,
				vpd_list_t ** vpd);

typedef char *
(node_get_parent_name_function_t) (const node_t     type,
				   const bus_info_t bus_info);

typedef char *
(node_get_name_function_t) (const node_t     type,
			    const bus_info_t bus_info);

typedef void
(node_set_yl_function_t) (const node_t     type,
			  const bus_info_t bus_info,
			  const char *     node,
			  vpd_list_t **    vpd);

typedef bool
(node_add_hook_function_t) (const node_t      type,
			    const bus_info_t  bus_info,
			    const node_t      subtype,
			    char **           node,
			    vpd_list_t **     vpd);

typedef char *
(node_get_source_dir_t) (const node_t     type,
			 const bus_info_t bus_info,
			 const node_t     subtype);

typedef node_t
(node_get_parent_subtype_t) (const node_t     bus_type,
			     const bus_addr_t bus_addr);

typedef struct node_functions {
	node_t                             type;
	node_t                             supertype;
	node_retrieve_vpd_function_t     * retrieve_vpd;
	node_get_parent_name_function_t  * get_parent_name;
	node_get_name_function_t         * get_kernel_name;
	node_get_name_function_t         * get_name;
	node_set_yl_function_t           * set_yl;   // FIXME: unused
	node_add_hook_function_t         * add_hook; // FIXME: unused
	node_get_source_dir_t            * get_source_dir;
	node_get_parent_subtype_t        * get_parent_subtype;
	struct node_functions            * next;
} node_functions_t;

extern node_functions_t * node_functions;

#define NODE_UNDEF ((void *) -1)
#define NODE_TYPE_DEFAULT "default"

/*
 * node_type_register:
 *
 * NULL fields in i are ignored.  Fields with value NODE_UNDEF cause
 * a previous value of that field for i->type to be NULLed out.  All
 * other non-NULL fields overwrite previous values for i->type.
 */
extern void
node_type_register(node_functions_t * i);

extern node_retrieve_vpd_function_t     node_retrieve_vpd;
extern node_get_parent_name_function_t  node_get_parent_name;
extern node_get_name_function_t         node_get_name;
extern node_get_name_function_t         node_get_kernel_name;
extern node_get_source_dir_t            node_get_source_dir;
extern node_get_parent_subtype_t        node_get_parent_subtype;

/********************************************************************/

/* Node listing. */

typedef bool (node_listing_function_t) (bus_info_list_t ** node_list);

/* This list is logically unordered. */
typedef struct node_listing_function_list {
	node_listing_function_t * f;
	struct node_listing_function_list * next;
} node_listing_function_list_t;

extern node_listing_function_list_t * node_listing_functions;

extern void
node_listing_functions_clear(void);

extern void
node_listing_function_register(node_listing_function_t * f);

extern bool
node_list_devices(bus_info_list_t ** node_list);

extern bool
node_list_add(bus_info_list_t ** node_list,
		const node_t type,
		const bus_info_t bus_info);

extern void
node_list_clear(bus_info_list_t ** node_list);

/********************************************************************/

/* Support functions. */

extern bool
bus_alias_set_type(const char * bus_info,
		   const char * target,
		   const char * alias_type);
extern bool
bus_alias_set(const char * bus_info,
	      const char * target);

extern bool
bus_alias_clear_type(const char * bus_info,
		     const char * alias_type);

extern bool
bus_alias_clear(const char * bus_info);

extern char *
bus_alias_get_type(const char * bus_info,
		   const char * alias_type);

extern char *
bus_alias_get(const char * bus_info);

/**/

extern char *
retrieve_node(const char * bus_info);

extern node_t
retrieve_subtype_basic(const char * node);

extern node_t
retrieve_subtype(const node_t bus_type,
		 const bus_addr_t bus_addr);

#endif
