/*
  sysfs - sysfs basics.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: sysfs.c,v 1.2 2013/09/14 04:41:14 munsone Exp $
*/

/* For asprintf. */
#define _GNU_SOURCE

#include <stdio.h>

// FIXME: Shouldn't be needed?
#include <linux/limits.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>

#include "util.h"
#include "init.h"
#include "sysfs.h"
#include "libsysfs.h"

char * lsvpd_sysfs_dir = NULL;

/********************************************************************/

char *
sysfs_get_device_node(const node_t bus_type,
		      const bus_addr_t bus_addr)
{
	char * ret;
	char * k;
	char * t;
	char buf[PATH_MAX+1];
	char * x;

	const char * sysfs_device_type = "sysfs-device";

	ret = NULL;

	k = lsvpd_relative_filename(bus_type, bus_addr);
	if (NULL != k)
		ret = bus_alias_get_type(k, sysfs_device_type);

	if (NULL == ret) {
		if (0 == strcmp(bus_type, BUS_TYPE_UNKNOWN)) {
			if (-1 == asprintf(&t, "%s/devices/%s",
					   lsvpd_sysfs_dir, bus_addr))
				t = NULL;
		} else {
			if (-1 == asprintf(&t, "%s/bus/%s/devices/%s",
					   lsvpd_sysfs_dir, bus_type, bus_addr))
				t = NULL;
		}

		if (NULL != t) {
			/* Not totally clear that realpath returns buf
			 * on success.  Also, realpath(3) says that
			 * passing a NULL value for buf causes
			 * realpath to malloc(3) a buffer and return
			 * it - however, this doesn't seemed to work
			 * all the time.  Perhaps it is a new feature
			 * or perhaps it isn't implemented on some
			 * architectures?  :-(  */
			x = realpath(t, buf);
			if (NULL != x)
				ret = strdup(x);
			free(t);
		}

		/* NOTE: If k isn't set it is annoying but not a
		 * showstopper so we still return the current value of
		 * ret.  Also, during the translation from the bash
		 * version, I've removed a sanity check that ensured
		 * ret started with /sys/devices/.  I don't think this
		 * is necessary anymore, but there may be a corner
		 * case somewhere.  */
		if ((NULL != ret) && (NULL != k))
			bus_alias_set_type(k, ret, sysfs_device_type);
	}

	free(k);

	return ret;
}

static char *
sysfs_get_source_dir(const node_t type,
		     bus_info_t   bus_info,
		     node_t       subtype)
{
	char * ret;
	node_t bus_type;
	bus_addr_t bus_addr;

	ret = NULL;

	if (bus_info_parse(bus_info, &bus_type, &bus_addr)) {
		ret = sysfs_get_device_node(bus_type, bus_addr);
		free(bus_type);
		free(bus_addr);
	}

	return ret;
}

/********************************************************************/

static void
init(void)
{
	char t[SYSFS_PATH_MAX];
	const char * x;
	node_functions_t n = {
		.type           = NODE_TYPE_DEFAULT,
		.get_source_dir = sysfs_get_source_dir,
	};

	x = getenv("LSVPD_SYSFS_DIR");
	if ((NULL == x) &&
	    (0 == sysfs_get_mnt_path(t, SYSFS_PATH_MAX)))
		x = t;

	if ((NULL != x) && (0 == access(x, F_OK))) {
		lsvpd_sysfs_dir = malloc(SYSFS_PATH_MAX + 1);
		if (NULL != lsvpd_sysfs_dir) {
			strncpy(lsvpd_sysfs_dir, x, SYSFS_PATH_MAX);
			lsvpd_sysfs_dir[SYSFS_PATH_MAX] = '\0';

			node_type_register(&n);
		}
	}
}

INIT(init);
